implement Hdict;

include "sys.m";
	sys: Sys;
	print, sprint: import sys;
include "draw.m";
include "arg.m";
include "bufio.m";
	bufio: Bufio;
	Iobuf: import bufio;
include "lists.m";
	lists: Lists;
include "lexicon.m";
	lexicon: Lexicon;
	Lex: import lexicon;

Hdict: module {
	init:	fn(nil: ref Draw->Context, nil: list of string);
};

init(nil: ref Draw->Context, args: list of string)
{
	sys = load Sys Sys->PATH;
	bufio = load Bufio Bufio->PATH;
	arg := load Arg Arg->PATH;
	lists = load Lists Lists->PATH;
	lexicon = load Lexicon Lexicon->PATH;
	if(lexicon == nil)
		fail(sprint("loading lexicon: %r"));

	addr := "dict.org";
	defaultstrat := "exact";
	defaultlex := "!";
	listlex := liststrats := serverinfo := 0;
	lexname: string;
	strat: string;
	infolex: string;
	verbose := quiet := 0;

	arg->init(args);
	arg->setusage(sprint("%s [-DIMvq] [-a addr] [-d lexicon] [-i lexicon] [-m strategy] [word]", arg->progname()));
	while((c := arg->opt()) != 0)
		case c {
		'a' =>	addr = arg->earg();
		'd' =>	lexname = arg->earg();
		'i' =>	infolex = arg->earg();
		'm' =>	strat = arg->earg();
		'D' =>	listlex = 1;
		'I' =>	serverinfo = 1;
		'M' =>	liststrats = 1;
		'v' =>	verbose = 1;
		'q' =>	quiet = 1;
		* =>	arg->usage();
		}

	args = arg->argv();
	if(len args > 1)
		arg->usage();

	if(lexname == nil && strat != nil)
		lexname = "*";
	if(strat == nil)
		strat = defaultstrat;
	if(lexname == nil)
		lexname = defaultlex;

	(lex, err) := lexicon->open(addr);
	if(err != nil)
		fail("connecting: "+err);
	lex.setdebug(verbose);

	first := 1;
	if(liststrats) {
		strats: list of (string, string);
		(strats, err) = lex.strategies();
		if(err != nil)
			fail("requesting strategies: "+err);

		if(!first)
			print("\n");
		first = 0;
		print("### Strategies\n");
		for(; strats != nil; strats = tl strats) {
			(name, descr) := hd strats;
			print("%-15s %s\n", name, descr);
		}
	}

	if(listlex) {
		lexs: list of (string, string);
		(lexs, err) = lex.lexicons();
		if(err != nil)
			fail("requesting lexicons: "+err);

		if(!first)
			print("\n");
		first = 0;
		print("### Lexicons\n");
		for(; lexs != nil; lexs = tl lexs) {
			(name, descr) := hd lexs;
			print("%-15s %s\n", name, descr);
		}
	}

	if(serverinfo) {
		text: string;
		(text, err) = lex.serverinfo();
		if(err != nil)
			fail("requesting serverinfo: "+err);

		if(!first)
			print("\n");
		first = 0;
		print("### Server information\n");
		print("%s\n", text);
	}

	if(infolex != nil) {
		text: string;
		(text, err) = lex.lexiconinfo(infolex);
		if(err != nil)
			fail("requesting lexiconinfo: "+err);

		if(!first)
			print("\n");
		first = 0;
		print("### Lexicon information about '%s'\n", infolex);
		print("%s\n", text);
	}

	if(len args == 0) {
		if(!liststrats && !listlex && !serverinfo && infolex == nil)
			arg->usage();
		return;
	}

	if(!first)
		print("\n");

	if(strat != defaultstrat) {
		words: list of (string, string);
		(words, err) = lex.match(hd args, lexname, strat);
		if(err != nil)
			fail("matching word: "+err);
		if(len words == 0)
			print("no matches\n");
		lexes: list of string;
	Nextword:
		for(l := words; l != nil; l = tl l) {
			(lname, nil) := hd l;
			for(m := lexes; m != nil; m = tl m)
				if(hd m == lname)
					continue Nextword;
			lexes = lname::lexes;
		}
		for(lexes = lists->reverse(lexes); lexes != nil; lexes = tl lexes) {
			for(l = words; l != nil; l = tl l) {
				(lname, word) := hd l;
				if(hd lexes != lname)
					continue;
				if(quiet)
					print("%q %q\n", lname, word);
				else
					print("hdict -d  %-8q %-15q\n", lname, word);
			}
		}
	} else {
		words: list of (string, string);
		err: string;
		if(lexname == defaultlex) {
			(words, err) = lex.match(hd args, lexname, strat);
			if(err != nil)
				fail("matching word: "+err);
		} else
			words = list of {(lexname, hd args)};

		have := 0;
		prevlex: string;
		for(nil; words != nil; words = tl words) {
			(rlex, rword) := hd words;
			defs: list of ref (string, string, string, string);
			(defs, err) = lex.define(rword, rlex);
			if(err != nil) {
				warn("error: "+err);
				continue;
			}
			if(len defs == 0)
				continue;
			have = 1;
			(nil, mlex, nil, nil) := *hd defs;
			if(rlex != lexname && rlex != prevlex)
				print("### From %s\n", mlex);
			prevlex = rlex;
			for(; defs != nil; defs = tl defs) {
				(nil, nil, nil, text) := *hd defs;
				print("%s\n", text);
				if(tl defs != nil)
					print("\n\n");
			}
		}
		if(!have)
			print("no matches\n");
	}
}

warn(s: string)
{
	sys->fprint(sys->fildes(2), "%s\n", s);
}

fail(s: string)
{
	warn(s);
	raise "fail:"+s;
}
