# a dict server procotol (rfc2229) implementation

# todo:
# - reconnect on failure

implement Lexicon;

include "sys.m";
	sys: Sys;
	sprint: import sys;
include "bufio.m";
	bufio: Bufio;
	Iobuf: import bufio;
include "dial.m";
	dial: Dial;
include "string.m";
	str: String;
include "lists.m";
	lists: Lists;
include "lexicon.m";


init()
{
	if(sys != nil)
		return;
	sys = load Sys Sys->PATH;
	bufio = load Bufio Bufio->PATH;
	dial = load Dial Dial->PATH;
	str = load String String->PATH;
	lists = load Lists Lists->PATH;
}

open(addr: string): (ref Lex, string)
{
	init();

	addr = dial->netmkaddr(addr, "net", "2628");
	c := dial->dial(addr, nil);
	if(c == nil)
		return (nil, sprint("dial %s: %r", addr));
	return fopen(c.dfd);
}

fopen(fd: ref Sys->FD): (ref Lex, string)
{
	init();

	{
		b := bufio->fopen(fd, bufio->OREAD);
		if(b == nil)
			error(sprint("bufio fopen: %r"));
		lx := ref Lex(fd, b, 0);

		l := read(lx);
		if(!str->prefix("220 ", l))
			error("could not handshake, response: "+l);
		code: string;
		(code, l) = command(lx, "CLIENT limbo dict client library");
		if(code != "250")
			return (nil, "error sending client version, response: "+l);
		return (lx, nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.lexicons(lx: self ref Lex): (list of (string, string), string)
{
	{
		(code, l) := command(lx, "SHOW DB");
		if(code != "110")
			error("show db failed, response: "+l);
		tuples := readtuples(lx);
		readokay(lx);
		return (tuples, nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.strategies(lx: self ref Lex): (list of (string, string), string)
{
	{
		(code, l) := command(lx, "SHOW STRAT");
		if(code != "111")
			error("show strat failed, response: "+l);
		tuples := readtuples(lx);
		readokay(lx);
		return (tuples, nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.serverinfo(lx: self ref Lex): (string, string)
{
	{
		(code, l) := command(lx, "SHOW SERVER");
		if(code != "114")
			error("show strat failed, response: "+l);
		text := readtext(lx);
		readokay(lx);
		return (text, nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.lexiconinfo(lx: self ref Lex, name: string): (string, string)
{
	{
		(code, l) := command(lx, sprint("SHOW INFO %s", quote(name)));
		if(code != "112")
			error("show info failed, response: "+l);
		text := readtext(lx);
		readokay(lx);
		return (text, nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.match(lx: self ref Lex, word, lexicon, strat: string): (list of (string, string), string)
{
	{
		if(lexicon == nil)
			lexicon = "!";
		if(strat == nil)
			strat = ".";

		if(!valid(word))
			error("invalid word: "+word);
		if(!valid(lexicon))
			error("invalid lexicon: "+lexicon);
		if(!valid(strat))
			error("invalid strategy: "+strat);

		(code, l) := command(lx, sprint("MATCH %s %s %s", quote(lexicon), quote(strat), quote(word)));
		if(code == "552")
			return (nil, nil);
		if(code[0] == '4' || code[0] == '5')
			error("match failed, response: "+l);

		tuples := readtuples(lx);
		readokay(lx);
		return (tuples, nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.define(lx: self ref Lex, word, lexicon: string): (list of ref (string, string, string, string), string)
{
	{
		if(lexicon == nil)
			lexicon = "!";

		if(!valid(word))
			error("invalid word: "+word);
		if(!valid(lexicon))
			error("invalid lexicon: "+lexicon);

		(code, l) := command(lx, sprint("DEFINE %s %s", quote(lexicon), quote(word)));
		if(code == "552")
			return (nil, nil);
		if(code[0] == '4' || code[0] == '5')
			error("define failed, response: "+l);

		r: list of ref (string, string, string, string);
		for(;;) {
			l = read(lx);
			if(!str->prefix("151 ", l))
				break;

			t := tokenize(l, 4);
			text := readtext(lx);
			r = ref (hd tl t, hd tl tl t, hd tl tl tl t, text)::r;
		}
		return (lists->reverse(r), nil);
	} exception e {
	"lexicon:*" =>
		return (nil, e[len "lexicon:":]);
	}
}

Lex.setdebug(lx: self ref Lex, debug: int)
{
	lx.debug = debug;
}


quote(s: string): string
{
	r: string;
	quote := 0;
	for(i := 0; i < len s; i++) {
		if(!quote && str->in(s[i], "a-zA-Z0-9"))
			quote = 1;
		if(str->in(s[i], "'\""))
			r += "'\\'";
		r += s[i:i+1];
	}
	if(quote)
		r = "'"+r+"'";
	return r;
}

readlist(lx: ref Lex): list of string
{
	r: list of string;
	for(;;) {
		l := read(lx);
		if(l == ".")
			break;
		if(str->prefix("..", l))
			l = l[1:];
		r = l::r;
	}
	return lists->reverse(r);
}

read(lx: ref Lex): string
{
	l := lx.b.gets('\n');
	if(l == "")
		error("end of file");

	if(lx.debug)
		sys->fprint(sys->fildes(2), "< %s", l);

	if(l[len l - 1] == '\n') {
		l = l[:len l-1];
		if(l[len l-1] == '\r')
			l = l[:len l-1];
	}
	return l;
}

readokay(lx: ref Lex)
{
	l := read(lx);
	if(!str->prefix("250 ", l))
		error("command not successful: "+l);
}

command(lx: ref Lex, cmd: string): (string, string)
{
	if(sys->fprint(lx.fd, "%s\r\n", cmd) < 0)
		error(sprint("writing command: %r"));

	if(lx.debug)
		sys->fprint(sys->fildes(2), "> %s\r\n", cmd);

	l := read(lx);
	if(len l < 3)
		error("invalid response: "+l);
	return (l[0:3], l);
	
}

readtuples(lx: ref Lex): list of (string, string)
{
	tokens: list of list of string;
	for(lines := readlist(lx); lines != nil; lines = tl lines)
		tokens = tokenize(hd lines, 2)::tokens;

	tuples: list of (string, string);
	for(; tokens != nil; tokens = tl tokens)
		tuples = (hd hd tokens, hd tl hd tokens)::tuples;
	return tuples;
}

readtext(lx: ref Lex): string
{
	s := "";
	for(lines := readlist(lx); lines != nil; lines = tl lines)
		s += "\n"+hd lines;
	if(s != nil)
		s = s[1:];
	return s;
}

valid(word: string): int
{
	if(word == nil)
		return 0;
	for(i := 0; i < len word; i++)
		if(word[i] < ' ')
			return 0;
	return 1;
}


tokenize(l: string, n: int): list of string
{
	r: list of string;
	origl := l;
	token: string;
	while(l != nil) {
		if(l[0] == '\'' || l[0] == '"') {
			quote := l[0];
			l = l[1:];
			for(;;) {
				if(l == nil)
					error("unbalanced quotes");
				if(l[0] == '\\') {
					if(len l == 1)
						error("missing value after escaping backslash, line: "+origl);
					token += l[1:2];
					l = l[2:];
				} else if(l[0] == quote) {
					if(len l > 1 && !str->in(l[1], " \t"))
						error("wrongly quoted tokens, line: "+origl);
					r = token::r;
					token = nil;

					l = l[1:];
					while(l != nil && str->in(l[0], " \t"))
						l = l[1:];
					break;
				} else {
					token += l[0:1];
					l = l[1:];
				}
			}
		} else if(l[0] == '\\') {
			if(len l == 1)
				error("missing value after escaping backslash, line: "+origl);
			r = l[1:2]::r;
			l = l[2:];
			if(len l > 1 && !str->in(l[0], " \t"))
				error("missing whitespace after escaping backslash, line: "+origl);
			while(l != nil && str->in(l[0], " \t"))
				l = l[1:];
		} else {
			(left, right) := str->splitl(l, " \t");
			r = left::r;
			if(right == nil)
				break;
			l = right;
			while(l != nil && str->in(l[0], " \t"))
				l = l[1:];
		}
	}
	if(len r != n)
		error("wrong number of tokens in response: "+origl);
	return lists->reverse(r);
}

error(s: string)
{
	raise "lexicon:"+s;
}
